package main

import (
	"context"
	"encoding/json"
	"flag"
	"fmt"
	"net"
	"os"
	"path"
	"sort"
	"strings"

	"github.com/sirupsen/logrus"

	"git.majava.org/software/dnsgen/pkg/config"
	"git.majava.org/software/dnsgen/pkg/dns"
	"git.majava.org/software/dnsgen/pkg/integration/kubernetes"
	"git.majava.org/software/dnsgen/pkg/integration/lookup"
	"git.majava.org/software/dnsgen/pkg/integration/netbox"
	"git.majava.org/software/dnsgen/pkg/integration/puppetdb"
	"git.majava.org/software/dnsgen/pkg/integration/webproxy"
)

var (
	debug               = flag.Bool("debug", false, "enable debug logging")
	configFilePath      = flag.String("config", "", "config file path")
	outputDirectoryPath = flag.String("output", "", "output directory")
)

func main() {
	flag.Parse()
	if *configFilePath == "" || *outputDirectoryPath == "" {
		logrus.Fatalln("--config or --output missing")
	}

	if *debug {
		logrus.SetLevel(logrus.DebugLevel)
	}

	var conf config.Config
	confData, err := os.ReadFile(*configFilePath)
	if err != nil {
		logrus.Fatalf("Failed to read config file: %v", err)
	}
	err = json.Unmarshal(confData, &conf)
	if err != nil {
		logrus.Fatalf("Failed to parse config: %v", err)
	}

	logrus.Debugf("%v", conf)

	perPrefixData := map[string][]dns.Record{}

	logrus.Infof("Now loading data from Netbox")
	nb := netbox.NewClient(conf.NetboxUrl, conf.NetboxToken)
	nbData, err := nb.FetchData(context.TODO())
	if err != nil {
		logrus.Fatalf("Failed to load data from netbox: %v", err)
	}

	for _, prefix := range conf.NetboxPrefixes {
		logrus.Infof("Now processing Netbox prefix %v", prefix)

		_, prefixNet, _ := net.ParseCIDR(prefix)
		netSize, _ := prefixNet.Mask.Size()
		subPrefixes := []*net.IPNet{}
		for _, subPrefixS := range conf.NetboxPrefixes {
			if subPrefixS == prefix {
				continue
			}

			subPrefixA, subPrefixNet, _ := net.ParseCIDR(subPrefixS)
			subPrefixNetSize, _ := subPrefixNet.Mask.Size()
			if prefixNet.Contains(subPrefixA) && subPrefixNetSize > netSize {
				subPrefixes = append(subPrefixes, subPrefixNet)
			}
		}

		records := []dns.Record{}

		addressRecords, err := nbData.GetIpAddressRecords(prefixNet, subPrefixes)
		if err != nil {
			logrus.Panicf("Failed to load Netbox address records: %v", err)
		}

		records = append(records, addressRecords...)

		rangeRecords, err := nbData.GetIpRangeRecords(prefixNet)
		if err != nil {
			logrus.Panicf("Failed to load Netbox range records: %v", err)
		}

		records = append(records, rangeRecords...)

		if _, ok := perPrefixData[prefix]; !ok {
			perPrefixData[prefix] = []dns.Record{}
		}

		perPrefixData[prefix] = append(perPrefixData[prefix], records...)
	}

	k8s := kubernetes.NewClient(conf.KubernetesConfig, conf.KubernetesPrefixes)

	for _, cluster := range conf.KubernetesClusters {
		logrus.Infof("Now processing K8s cluster %v", cluster)

		recordsPerPrefix, err := k8s.GetRecords(cluster)
		if err != nil {
			logrus.Panicf("Failed to load Kubernetes records for cluster %v: %v", cluster, err)
		}

		for prefix, records := range recordsPerPrefix {
			if _, ok := perPrefixData[prefix]; !ok {
				perPrefixData[prefix] = []dns.Record{}
			}

			perPrefixData[prefix] = append(perPrefixData[prefix], records...)
		}
	}

	for _, query := range conf.PuppetdbQueries {
		logrus.Infof("Now doing PuppetDB query %v", query.Name)
		client := puppetdb.NewClient(query.Connection)
		prefix := fmt.Sprintf("pdb-%s", query.Name)
		perPrefixData[prefix] = []dns.Record{}

		for _, recordQuery := range query.Records {
			records, err := client.GetRecords(recordQuery)
			if err != nil {
				logrus.Panicf("Failed to load %v records for PuppetDB: %v", query.Name, err)
			}

			perPrefixData[prefix] = append(perPrefixData[prefix], records...)
		}
	}

	for _, proxy := range conf.WebProxies {
		logrus.Infof("Now processing web proxy %v", proxy.Name)
		records, err := webproxy.GetRecords(proxy)
		if err != nil {
			logrus.Panicf("Failed to load proxy records for proxy %v: %v", proxy.Name, err)
		}

		perPrefixData[fmt.Sprintf("webproxy-%v", proxy.Name)] = records
	}

	for _, lookupConf := range conf.Lookups {
		logrus.Infof("Now processing lookup %v", lookupConf.Name)
		records, err := lookup.GetRecords(lookupConf)
		if err != nil {
			logrus.Panicf("Failed to load lookup records for %v: %v", lookupConf.Name, err)
		}

		logrus.Debugf("lookup records = %v", records)
		perPrefixData[fmt.Sprintf("lookup-%v", lookupConf.Name)] = records
	}

	logrus.Infof("Now exporting zone files")

	sort.Slice(conf.Domains, func(i1, i2 int) bool {
		return strings.Count(conf.Domains[i1], ".") > strings.Count(conf.Domains[i2], ".")
	})

	for prefix, records := range perPrefixData {
		prefixDir := path.Join(*outputDirectoryPath, strings.ReplaceAll(prefix, "/", "-"))
		err := os.Mkdir(prefixDir, 0o0755)
		if err != nil {
			logrus.Fatalf("Failed to create directory %v, %v", prefixDir, err)
		}

		for _, domain := range conf.Domains {
			subdomains := []string{}
			for _, possibleSubdomain := range conf.Domains {
				if domain == possibleSubdomain {
					continue
				}

				if strings.HasSuffix(possibleSubdomain, "."+domain) {
					subdomains = append(subdomains, possibleSubdomain)
				}
			}

			zoneFile := dns.ZoneFile{
				Zone:    domain,
				Records: []dns.Record{},
			}

			for _, record := range records {
				if !strings.HasSuffix("."+record.Name, "."+domain) {
					continue
				}

				foundSubdomain := false
				for _, possibleSubdomain := range subdomains {
					if strings.HasSuffix("."+record.Name, "."+possibleSubdomain) {
						foundSubdomain = true
						break
					}
				}
				if foundSubdomain {
					continue
				}

				if record.Type == dns.RecordTypeCNAME && record.Name == zoneFile.Zone {
					logrus.Warnf("Skipping apex CNAME record %s (%s)", record.Name, record.Value)
					continue
				}

				zoneFile.Records = append(zoneFile.Records, record)
			}

			file := zoneFile.Render()
			if file != "" {
				err := os.WriteFile(path.Join(prefixDir, domain+"zone"), []byte(file), 0o0644)
				if err != nil {
					logrus.Fatalf("Failed to write %v/%v %v", prefixDir, domain, err)
				}
			}
		}
	}

	logrus.Infof("All done.")
}
