package dns

import (
	"slices"
	"strings"

	"facette.io/natsort"
	miekgdns "github.com/miekg/dns"
)

type RecordType string

const (
	RecordTypeA     RecordType = "A"
	RecordTypeAAAA  RecordType = "AAAA"
	RecordTypeCNAME RecordType = "CNAME"
	RecordTypeMX    RecordType = "MX"
	RecordTypePTR   RecordType = "PTR"
	RecordTypeTXT   RecordType = "TXT"
)

const LongestRecordTypeLength = 5

var (
	recordTypeMap = map[string]RecordType{
		"a":     RecordTypeA,
		"aaaa":  RecordTypeAAAA,
		"cname": RecordTypeCNAME,
		"mx":    RecordTypeMX,
		"ptr":   RecordTypePTR,
		"txt":   RecordTypeTXT,
	}

	recordTypeNumberMap = map[RecordType]uint16{
		RecordTypeA:     miekgdns.TypeA,
		RecordTypeAAAA:  miekgdns.TypeAAAA,
		RecordTypeCNAME: miekgdns.TypeCNAME,
		RecordTypeMX:    miekgdns.TypeMX,
		RecordTypePTR:   miekgdns.TypePTR,
		RecordTypeTXT:   miekgdns.TypeTXT,
	}
)

func (t RecordType) Name() string {
	return string(t)
}

func (t RecordType) ForQuery() uint16 {
	return recordTypeNumberMap[t]
}

func RecordTypeFor(s string) (RecordType, bool) {
	t, ok := recordTypeMap[strings.ToLower(s)]
	return t, ok
}

type Record struct {
	Name  string
	Type  RecordType
	Value string
}

// Compare compares this Record to some other Record.
func (r Record) Compare(that Record) bool {
	partsThis := strings.Split(r.Name, ".")
	partsThat := strings.Split(that.Name, ".")
	slices.Reverse(partsThis)
	slices.Reverse(partsThat)

	lenThis := len(partsThis)
	lenThat := len(partsThat)

	for i := 0; i < lenThis && i < lenThat; i++ {
		if partsThis[i] != partsThat[i] {
			return natsort.Compare(partsThis[i], partsThat[i])
		}

		if i+1 == lenThis && i+1 != lenThat {
			return true
		}
		if i+1 == lenThat && i+1 != lenThis {
			return false
		}
	}

	if r.Type != that.Type {
		return natsort.Compare(r.Type.Name(), that.Type.Name())
	}

	return natsort.Compare(r.Value, that.Value)
}
