package kubernetes

import (
	"context"
	"errors"
	"fmt"
	"net"

	"github.com/sirupsen/logrus"
	corev1 "k8s.io/api/core/v1"
	metav1 "k8s.io/apimachinery/pkg/apis/meta/v1"
	k8s "k8s.io/client-go/kubernetes"
	"k8s.io/client-go/tools/clientcmd"

	"git.majava.org/software/dnsgen/pkg/dns"
)

type Client struct {
	configFile        string
	prefixes          []string
	ingressClassCache map[string]map[string]*string
}

func NewClient(configFile string, prefixes []string) Client {
	return Client{
		configFile:        configFile,
		prefixes:          prefixes,
		ingressClassCache: map[string]map[string]*string{},
	}
}

func (client *Client) getIngressClassFrontProxy(c *k8s.Clientset, cluster string, ingressClass string) (*string, error) {
	clusterCache, ok := client.ingressClassCache[cluster]
	if !ok {
		clusterCache = map[string]*string{}
		client.ingressClassCache[cluster] = clusterCache
	}

	if cachedProxy, ok := clusterCache[ingressClass]; ok {
		return cachedProxy, nil
	}

	logrus.Debugf("k8s: Quering front proxy data for ingress class %v on cluster %v", ingressClass, cluster)

	class, err := c.NetworkingV1().IngressClasses().Get(context.TODO(), ingressClass, metav1.GetOptions{})
	if err != nil {
		return nil, err
	}

	if class == nil {
		return nil, errors.New(fmt.Sprintf("Could not find ingress class %v from cluster %v", ingressClass, cluster))
	}

	annotation, ok := class.ObjectMeta.Annotations["dnsgen.majava.org/front-proxy"]
	if !ok {
		client.ingressClassCache[cluster][ingressClass] = nil
		return nil, nil
	}

	client.ingressClassCache[cluster][ingressClass] = &annotation
	return &annotation, nil
}

func (client *Client) generateRecords(
	c *k8s.Clientset,
	results map[string][]dns.Record,
	host string, ip string,
	cluster string,
	ingressClass *string,
) (map[string][]dns.Record, error) {
	if ip == "" {
		return results, nil
	}

	prefix := ""
	value := ""
	reverseName := ""
	recordType := dns.RecordTypeAAAA

	if ingressClass != nil {
		frontProxy, err := client.getIngressClassFrontProxy(c, cluster, *ingressClass)
		if err != nil {
			return results, err
		}

		if frontProxy != nil {
			prefix = fmt.Sprintf("kube-%v", cluster)
			recordType = dns.RecordTypeCNAME
			value = *frontProxy
		}
	}

	if prefix == "" {
		addr := net.ParseIP(ip)
		if addr == nil {
			return results, nil
		}

		if addr.To4() != nil {
			recordType = dns.RecordTypeA
		}

		value = addr.String()
		reverseName = dns.GetReverseName(addr, recordType)

		prefix = dns.FindMostSpecificPrefix(client.prefixes, addr)
		if prefix == "" {
			return results, nil
		}
	}

	if _, ok := results[prefix]; !ok {
		results[prefix] = []dns.Record{}
	}

	forwardRecord := dns.Record{
		Name:  host + ".",
		Type:  recordType,
		Value: value,
	}

	results[prefix] = append(results[prefix], forwardRecord)

	if reverseName != "" {
		reverseRecord := dns.Record{
			Name:  reverseName,
			Type:  dns.RecordTypePTR,
			Value: host + ".",
		}

		results[prefix] = append(results[prefix], reverseRecord)
	}

	return results, nil
}

func (client *Client) GetRecords(cluster string) (map[string][]dns.Record, error) {
	config, err := clientcmd.NewNonInteractiveDeferredLoadingClientConfig(
		&clientcmd.ClientConfigLoadingRules{ExplicitPath: client.configFile},
		&clientcmd.ConfigOverrides{CurrentContext: cluster}).ClientConfig()
	if err != nil {
		return nil, err
	}

	c, err := k8s.NewForConfig(config)
	if err != nil {
		return nil, err
	}

	results := map[string][]dns.Record{}

	services, err := c.CoreV1().
		Services(metav1.NamespaceAll).
		List(context.TODO(), metav1.ListOptions{})
	if err != nil {
		return nil, err
	}

	for _, svc := range services.Items {
		if svc.Spec.Type != corev1.ServiceTypeLoadBalancer {
			continue
		}

		annotation, ok := svc.ObjectMeta.Annotations["dnsgen.majava.org/external-name"]
		if !ok {
			continue
		}

		for _, ing := range svc.Status.LoadBalancer.Ingress {
			results, err = client.generateRecords(c, results, annotation, ing.IP, cluster, nil)
			if err != nil {
				return nil, err
			}
		}
	}

	ingresses, err := c.NetworkingV1().
		Ingresses(metav1.NamespaceAll).
		List(context.TODO(), metav1.ListOptions{})
	if err != nil {
		return nil, err
	}

	for _, ing := range ingresses.Items {
		hosts := map[string]bool{}
		for _, rule := range ing.Spec.Rules {
			hosts[rule.Host] = true
		}

		for _, ingStatus := range ing.Status.LoadBalancer.Ingress {
			for host := range hosts {
				results, err = client.generateRecords(c, results, host, ingStatus.IP, cluster, ing.Spec.IngressClassName)
				if err != nil {
					return nil, err
				}
			}
		}
	}

	return results, nil
}
