// Package lookup provides an integration that simply queries a configured
// DNS record.
package lookup

import (
	"errors"
	"fmt"
	"net"
	"time"

	"github.com/sirupsen/logrus"

	"git.majava.org/software/dnsgen/pkg/dns"

	miekgdns "github.com/miekg/dns"
)

type Query struct {
	Source string   `json:"source"`
	Target string   `json:"target"`
	Types  []string `json:"types"`
}

type Lookup struct {
	Name    string  `json:"name"`
	Queries []Query `json:"queries"`
}

func parseResult(rr miekgdns.RR, recordType dns.RecordType) string {
	switch recordType {
	case dns.RecordTypeA:
		return rr.(*miekgdns.A).A.String()
	case dns.RecordTypeAAAA:
		return rr.(*miekgdns.AAAA).AAAA.String()
	default:
		panic(fmt.Sprintf("unsupported type %v", recordType))
	}
}

func getQueryRecords(query Query, c *miekgdns.Client, nameserver string) ([]dns.Record, error) {
	records := []dns.Record{}
	sourceFqdn := miekgdns.Fqdn(query.Source)
	targetFqdn := miekgdns.Fqdn(query.Target)

	for _, recordTypeName := range query.Types {
		recordType, ok := dns.RecordTypeFor(recordTypeName)
		if !ok {
			logrus.Warnf("Ignoring request for %s => %s with invalid type %s", query.Source, query.Target, recordTypeName)
			continue
		}

		msg := &miekgdns.Msg{}
		msg.SetQuestion(sourceFqdn, recordType.ForQuery())
		msg.RecursionDesired = true

		r, _, err := c.Exchange(msg, nameserver)
		if err != nil {
			return nil, err
		}

		for _, rr := range r.Answer {
			records = append(records, dns.Record{
				Name:  targetFqdn,
				Type:  recordType,
				Value: parseResult(rr, recordType),
			})
		}
	}

	if len(records) == 0 {
		return nil, errors.New(fmt.Sprintf("Got no records for query %v", query))
	}
	return records, nil
}

func GetRecords(config Lookup) ([]dns.Record, error) {
	records := []dns.Record{}

	dnsConfig, err := miekgdns.ClientConfigFromFile("/etc/resolv.conf")
	if err != nil {
		return nil, err
	}
	nameserver := net.JoinHostPort(dnsConfig.Servers[0], dnsConfig.Port)

	c := &miekgdns.Client{
		Timeout: 1 * time.Second,
	}

	for _, query := range config.Queries {
		queryRecords, err := getQueryRecords(query, c, nameserver)
		if err != nil {
			return nil, err
		}

		records = append(records, queryRecords...)
	}

	return records, err
}
