package netbox

import (
	"bytes"
	"context"
	"fmt"
	"net"
	"net/http"
	"strings"
	"time"

	"git.sr.ht/~emersion/gqlclient"
	"github.com/hashicorp/go-retryablehttp"

	"git.majava.org/software/dnsgen/pkg/dns"
)

type headerTransport struct {
	apiToken string
	r        http.RoundTripper
}

func (t headerTransport) RoundTrip(r *http.Request) (*http.Response, error) {
	r.Header.Add("Authorization", fmt.Sprintf("token %v", t.apiToken))
	return t.r.RoundTrip(r)
}

type Client struct {
	gql *gqlclient.Client
}

// Data is the set of cached data loaded from Netbox
type Data struct {
	ips    []IPAddressType
	ranges []IPRangeType
}

func NewClient(endpoint, apiToken string) *Client {
	retryClient := retryablehttp.NewClient()
	retryClient.Logger = hclogrus{}
	retryClient.RetryMax = 5
	retryClient.HTTPClient.Timeout = time.Second * 10

	httpClient := retryClient.StandardClient()
	httpClient.Transport = headerTransport{
		r:        httpClient.Transport,
		apiToken: apiToken,
	}

	return &Client{
		gql: gqlclient.New(endpoint, httpClient),
	}
}

func (c *Client) FetchData(ctx context.Context) (*Data, error) {
	ips, err := ListIps(c.gql, ctx)
	if err != nil {
		return nil, err
	}

	ranges, err := ListIpRanges(c.gql, context.TODO())
	if err != nil {
		return nil, err
	}

	return &Data{
		ips:    ips,
		ranges: ranges,
	}, nil
}

func (d *Data) GetIpAddressRecords(prefix *net.IPNet, subPrefixes []*net.IPNet) ([]dns.Record, error) {
	records := []dns.Record{}

	for _, ip := range d.ips {
		if ip.Dns_name == "" {
			continue
		}

		if ip.Status != "active" && ip.Status != "deprecated" {
			continue
		}

		addr, _, _ := net.ParseCIDR(ip.Address)

		if !prefix.Contains(addr) {
			continue
		}

		process := true
		for _, subPrefix := range subPrefixes {
			if subPrefix.Contains(addr) {
				process = false
			}
		}
		if !process {
			continue
		}

		recordType := dns.RecordTypeAAAA
		if addr.To4() != nil {
			recordType = dns.RecordTypeA
		}

		forwardRecord := dns.Record{
			Name:  ip.Dns_name + ".",
			Type:  recordType,
			Value: addr.String(),
		}

		reverseRecord := dns.Record{
			Name:  dns.GetReverseName(addr, recordType),
			Type:  dns.RecordTypePTR,
			Value: ip.Dns_name + ".",
		}

		records = append(records, forwardRecord, reverseRecord)
	}

	return records, nil
}

func (d *Data) GetIpRangeRecords(prefix *net.IPNet) ([]dns.Record, error) {
	records := []dns.Record{}

	for _, ipRange := range d.ranges {
		if ipRange.Role.Name != "DHCP" {
			continue
		}

		if ipRange.Status != "active" && ipRange.Status != "deprecated" {
			continue
		}

		startAddress, startNet, _ := net.ParseCIDR(ipRange.Start_address)
		if startNet.String() != prefix.String() {
			continue
		}

		dnsTemplate, ok := ipRange.Custom_fields["dns_template"].(string)
		if !ok {
			continue
		}

		recordType := dns.RecordTypeAAAA
		if startAddress.To4() != nil {
			recordType = dns.RecordTypeA
		}

		endAddress, _, _ := net.ParseCIDR(ipRange.End_address)

		current := startAddress

		for {
			// TODO: ipv6 support
			asString := strings.ReplaceAll(current.String(), ".", "-")
			nameForCurrent := strings.ReplaceAll(dnsTemplate, "$1", asString)

			forwardRecord := dns.Record{
				Name:  nameForCurrent,
				Type:  recordType,
				Value: current.String(),
			}

			reverseRecord := dns.Record{
				Name:  dns.GetReverseName(current, recordType),
				Type:  dns.RecordTypePTR,
				Value: nameForCurrent,
			}

			records = append(records, forwardRecord, reverseRecord)

			current = dns.NextAddress(current, 1)
			if bytes.Compare(current, endAddress) > 0 {
				break
			}
		}
	}

	return records, nil
}
