package puppetdb

import (
	"errors"
	"fmt"

	"github.com/Jeffail/gabs"
	"github.com/akira/go-puppetdb"
	"github.com/sirupsen/logrus"

	"git.majava.org/software/dnsgen/pkg/dns"
)

type ClientConfig struct {
	Host          string `json:"host"`
	Port          int    `json:"port"`
	ClientCert    string `json:"client_cert"`
	ClientCertKey string `json:"client_cert_key"`
	CaCert        string `json:"ca_cert"`
}

type Query struct {
	Category  string `json:"category"`
	HostFqdn  string `json:"host_fqdn"`
	Interface string `json:"interface"`
	DnsName   string `json:"dns_name"`
	OnlyProto int    `json:"only_proto"`
}

type Client struct {
	client *puppetdb.Client
}

func NewClient(conf ClientConfig) Client {
	if conf.ClientCert == "" {
		return Client{
			client: puppetdb.NewClientTimeout(
				conf.Host,
				conf.Port,
				false,
				15,
			),
		}
	}
	return Client{
		client: puppetdb.NewClientTimeoutSSL(
			conf.Host,
			conf.Port,
			conf.ClientCertKey,
			conf.ClientCert,
			conf.CaCert,
			false,
			15,
		),
	}
}

func (c Client) getFactForHost(host string, fact string) (*gabs.Container, error) {
	facts, err := c.client.NodeFacts(host)
	if err != nil {
		return nil, err
	}

	for _, f := range facts {
		if f.Name == fact {
			return f.Value, nil
		}
	}

	return nil, errors.New("fact not found")
}

func (c Client) getCategoryRecords(query Query) ([]dns.Record, error) {
	records := []dns.Record{}

	result, err := c.client.Resources(
		fmt.Sprintf(
			"[\"and\", [\"=\", \"type\", \"Dnsgen::Export\"], [\"=\", [\"parameter\", \"category\"], \"%v\"]]",
			query.Category,
		),
		map[string]string{},
	)

	if err != nil {
		return nil, err
	}

	for _, res := range result {
		recordType, ok := dns.RecordTypeFor(res.Paramaters["type"].(string))
		if !ok {
			logrus.Warnf("Ignoring record %s from %s with invalid type %s", res.Title, res.Certname, res.Paramaters["type"])
			continue
		}

		records = append(records, dns.Record{
			Name:  res.Paramaters["fqdn"].(string),
			Type:  recordType,
			Value: res.Paramaters["contents"].(string),
		})
	}

	return records, nil
}

func (c Client) getHostRecords(query Query) ([]dns.Record, error) {
	records := []dns.Record{}

	fact, err := c.getFactForHost(query.HostFqdn, "networking")
	if err != nil {
		return nil, err
	}

	iface := fact.Path("interfaces").Path(query.Interface)

	if iface.Exists("ip") && query.OnlyProto != 6 {
		records = append(records, dns.Record{
			Name:  query.DnsName,
			Type:  dns.RecordTypeA,
			Value: iface.Path("ip").Data().(string),
		})
	}
	if iface.Exists("ip6") && query.OnlyProto != 4 {
		records = append(records, dns.Record{
			Name:  query.DnsName,
			Type:  dns.RecordTypeAAAA,
			Value: iface.Path("ip6").Data().(string),
		})
	}

	return records, nil
}

func (c Client) GetRecords(query Query) ([]dns.Record, error) {
	if query.Category != "" {
		return c.getCategoryRecords(query)
	}

	if query.HostFqdn != "" {
		return c.getHostRecords(query)
	}

	return nil, errors.New("unsupported query type")
}
